<?php
use Drupal\Core\Config\FileStorage;
use Drush\Log\LogLevel;
use Drupal\Core\Config\StorageInterface;
use Drupal\Core\Config\StorageComparer;

/**
 * @file
 * Drush integration for the config_partial_export module.
 */

/**
 * Implements hook_drush_command().
 */
function config_partial_export_drush_command() {
  $items['config-partial-export'] = array(
    'description' => 'Export a partial configuration to a directory.',
    'core' => array('8+'),
    'aliases' => array('cpex'),
    'arguments' => array(
      'config' => "Configuration keys, comma separated",
      'label' => "A config directory label (i.e. a key in \$config_directories array in settings.php). Defaults to 'sync'",
    ),
    'options' => array(
      'changelist' => 'Shows the list of changed active config.'
    ),
    'required-arguments' => 1,
    'examples' => array(
      'drush config-partial-export --config core.date_format.long' => 'Export configuration; Save files in a backup directory named config-export.',
    ),
  );
  return $items;
}

/**
 * Drush command callback to perform a partial config export.
 *
 * @param $config
 * @return mixed
 */
function drush_config_partial_export($config) {
  $changelist = drush_get_option('changelist', 0);
  if ($changelist) {
    $changes = _config_partial_export_get_changes();
    if (!empty($changes)) {
      drush_print(dt("Your configuration has changed:"));
      foreach ($changes as $key => $value) {
        drush_print($value);
      }
    }
    return TRUE;
  }
  global $config_directories;
  $choices = drush_map_assoc(array_keys($config_directories));
  unset($choices[CONFIG_ACTIVE_DIRECTORY]);
  if (!isset($destination) && count($choices) >= 2) {
    $destination = drush_choice($choices, 'Choose a destination.');
    if (empty($destination)) {
      return drush_user_abort();
    }
  }
  elseif (!isset($destination)) {
    $destination = CONFIG_SYNC_DIRECTORY;
  }
  $destination_dir = config_get_config_directory($destination);
  $destination_storage = new FileStorage($destination_dir);
  $source_storage = \Drupal::service('config.storage');

  $config_keys = explode(",", $config);
  foreach ($config_keys as $config_key) {
    // Look for a wildcard character.
    if (strpos($config_key, '*') !== FALSE) {
      $wildcard_keys = _config_partial_export_get_wildcard_keys($config_key, $source_storage);
      foreach ($wildcard_keys as $wildcard_key) {
        _config_partial_export_write_config($wildcard_key, $source_storage, $destination_storage, $destination_dir);
      }
    }
    else {
      _config_partial_export_write_config($config_key, $source_storage, $destination_storage, $destination_dir);
    }
  }

  return TRUE;
}

/**
 * Writes a YAML configuration file to the specified directory.
 *
 * @param string $key
 *   Configuration key.
 * @param \Drupal\Core\Config\StorageInterface $source_storage
 *   The source storage.
 * @param \Drupal\Core\Config\StorageInterface $destination_storage
 *   The source storage.
 *
 * @return bool
 *   Whether or not the configuration was moved from source to destination.
 */
function _config_partial_export_write_config($key, StorageInterface $source_storage, StorageInterface $destination_storage, $destination_dir) {
  $config_exists = $source_storage->read($key);

  if ($config_exists) {
    $destination_storage->write($key, $config_exists);
    drush_log(dt('Writing !name to !target.', array(
      '!name' => $key,
      '!target' => $destination_dir,
    )), LogLevel::SUCCESS);
  }
  else {
//    drush_set_error('config_name_not_found', dt('!name not found in the configuration.', array('!name' => $key)));
    drush_log(dt('!name not found in the configuration.', array('!name' => $key)), LogLevel::ERROR);
  }

  return $config_exists;
}

/**
 * Checking if a configuration matches a wildcard.
 *
 * @param string $input
 *   The string that contains the wildcard.
 * @param \Drupal\Core\Config\StorageInterface $storage
 *   The source storage.
 *
 * @return array
 *   The list of keys.
 */
function _config_partial_export_get_wildcard_keys($input, StorageInterface $storage) {
  // Get the strings around the wildcard.
  $split = explode('*', $input);
  $matching_keys = [];

  // Load the possible keys that start with the first prefix.
  $possible_keys = $storage->listAll($split[0]);

  // Check each key if they match the strings.
  foreach ($possible_keys as $config_key) {
    $match = TRUE;
    $counter = strlen($split[0]);

    for ($i = 1; $i < count($split); $i++) {
      if (!empty($split[$i])) {
        // Check if the partial exists after the last check.
        $pos = strpos($config_key, $split[$i], $counter);
        // If no "match" was found for this partial, it should fail.
        if ($pos === FALSE) {
          $match = FALSE;
        }
        // Increment the counter by the position found and length of the match.
        $counter += ($pos + strlen($split[$i]));
      }
    }
    if ($match) {
      $matching_keys[] = $config_key;
    }
  }
  return $matching_keys;
}

/**
 * Gets the list of changed configurations.
 *
 * @return array|bool
 *   TRUE if there are no changes.
 */
function _config_partial_export_get_changes() {
  $sync_storage = \Drupal::service('config.storage.sync');
  $active_storage = \Drupal::service('config.storage');
  $config_manager = \Drupal::service('config.manager');

  $storage_comparer = new StorageComparer($sync_storage, $active_storage, $config_manager);

  $source_list = $sync_storage->listAll();
  $change_list = $storage_comparer->createChangelist();
  if (empty($source_list) || !$change_list->hasChanges()) {
    drush_print(dt('There are no configuration changes.'));
    return TRUE;
  }
  $result = array();
  if (!empty($change_list) && $change_list->hasChanges()) {
    foreach ($change_list->getAllCollectionNames() as $collection) {
      foreach ($change_list->getChangelist(NULL, $collection) as $action => $config_names) {
        if (empty($config_names)) {
          continue;
        }
        foreach ($config_names as $config_name) {
          $result[] = $config_name;
        }
      }
    }
    sort($result);
  }

  return $result;
}
